/*
Copyright (c) 2019-2022 Digital Energy Cloud Solutions LLC. All Rights Reserved.
Authors:
Petr Krutov, <petr.krutov@digitalenergy.online>
Stanislav Solovev, <spsolovev@digitalenergy.online>
Kasim Baybikov, <kmbaybikov@basistech.ru>
Tim Tkachev, <tvtkachev@basistech.ru>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
Terraform DECORT provider - manage resources provided by DECORT (Digital Energy Cloud
Orchestration Technology) with Terraform by Hashicorp.

Source code: https://repository.basistech.ru/BASIS/terraform-provider-decort

Please see README.md to learn where to place source code so that it
builds seamlessly.

Documentation: https://repository.basistech.ru/BASIS/terraform-provider-decort/wiki
*/

package extnet

import (
	"context"
	"strconv"

	"github.com/hashicorp/terraform-plugin-sdk/v2/diag"
	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/schema"
	log "github.com/sirupsen/logrus"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/extnet"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/constants"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/controller"
)

func resourceExtnetCreate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("cloudbroker: resourceExtnetCreate called with name '%s'", d.Get("name").(string))
	c := m.(*controller.ControllerCfg)

	req := extnet.CreateRequest{
		Name:   d.Get("name").(string),
		GID:    uint64(d.Get("gid").(int)),
		IPCIDR: d.Get("ipcidr").(string),
		VLANID: uint64(d.Get("vlan_id").(int)),
	}

	if gateway, ok := d.GetOk("gateway"); ok {
		req.Gateway = gateway.(string)
	}

	if dns, ok := d.GetOk("dns"); ok {
		res := make([]string, 0)
		for _, elem := range dns.([]interface{}) {
			res = append(res, elem.(string))
		}

		req.DNS = res
	}

	if ntp, ok := d.GetOk("ntp"); ok {
		res := make([]string, 0)
		for _, elem := range ntp.([]interface{}) {
			res = append(res, elem.(string))
		}

		req.NTP = res
	}

	if check_ips, ok := d.GetOk("check_ips"); ok {
		res := make([]string, 0)
		for _, elem := range check_ips.([]interface{}) {
			res = append(res, elem.(string))
		}

		req.CheckIPs = res
	}

	if virtual, ok := d.GetOk("virtual"); ok {
		req.Virtual = virtual.(bool)
	}

	if desc, ok := d.GetOk("desc"); ok {
		req.Description = desc.(string)
	}

	if start_ip, ok := d.GetOk("start_ip"); ok {
		req.StartIP = start_ip.(string)
	}

	if end_ip, ok := d.GetOk("end_ip"); ok {
		req.EndIP = end_ip.(string)
	}

	if vnfdev_ip, ok := d.GetOk("vnfdev_ip"); ok {
		req.VNFDevIP = vnfdev_ip.(string)
	}

	if pre_reservations_num, ok := d.GetOk("pre_reservations_num"); ok {
		req.PreReservationsNum = uint64(pre_reservations_num.(int))
	}

	if ovs_bridge, ok := d.GetOk("ovs_bridge"); ok {
		req.OVSBridge = ovs_bridge.(string)
	}

	log.Debugf("cloudbroker: Sent create request")
	netID, err := c.CloudBroker().ExtNet().Create(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	d.SetId(strconv.FormatUint(netID, 10))
	log.Debugf("cloudbroker: Extnet with id %d successfully created on platform", netID)

	if d.Get("excluded_ips").(*schema.Set).Len() > 0 {
		ips := make([]string, 0)

		for _, ip := range d.Get("excluded_ips").(*schema.Set).List() {
			ips = append(ips, ip.(string))
		}

		req := extnet.IPsExcludeRequest{
			NetID: netID,
			IPs:   ips,
		}

		_, err := c.CloudBroker().ExtNet().IPsExclude(ctx, req)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	return resourceExtnetRead(ctx, d, m)
}

func resourceExtnetRead(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("cloudbroker: resourceExtnetRead called with id %s", d.Id())
	recNet, err := utilityExtnetCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	flattenRecordExtnet(d, recNet)

	return nil
}

func resourceExtnetUpdate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("cloudbroker: resourceExtnetUpdate called with id %s", d.Id())
	c := m.(*controller.ControllerCfg)

	recNet, err := utilityExtnetCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	if err := handleBasicUpdate(ctx, d, c, recNet); err != nil {
		return diag.FromErr(err)
	}

	if d.HasChange("enable") {
		if err := handleEnableUpdate(ctx, d, c, recNet); err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("default_qos") {
		if err := handleDefaultQOSUpdate(ctx, d, c, recNet); err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("dns") {
		if err := handleDNSUpdate(ctx, d, c, recNet); err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("ntp") {
		if err := handleNTPUpdate(ctx, d, c, recNet); err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("set_default") {
		if err := handleSetDefault(ctx, d, c, recNet); err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("excluded_ips") {
		if err := handleExcludedIPsUpdate(ctx, d, c, recNet); err != nil {
			return diag.FromErr(err)
		}
	}

	return resourceExtnetRead(ctx, d, m)
}

func resourceExtnetDelete(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("cloudbroker: resourceExtnetDelete called with id %s", d.Id())
	c := m.(*controller.ControllerCfg)

	netId, _ := strconv.ParseUint(d.Id(), 10, 64)
	req := extnet.DestroyRequest{NetID: netId}

	_, err := c.CloudBroker().ExtNet().Destroy(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}
	log.Debugf("cloudbroker: successfully destroyed extnet with id %s", d.Id())

	return nil
}

func resourceExtnetSchemaMake() map[string]*schema.Schema {
	return map[string]*schema.Schema{
		"name": {
			Type:        schema.TypeString,
			Required:    true,
			Description: "External network name",
		},
		"gid": {
			Type:        schema.TypeInt,
			Required:    true,
			Description: "Grid (platform) ID",
		},
		"ipcidr": {
			Type:        schema.TypeString,
			Required:    true,
			ForceNew:    true,
			Description: "IP network CIDR",
		},
		"vlan_id": {
			Type:        schema.TypeInt,
			Required:    true,
			ForceNew:    true,
			Description: "VLAN ID",
		},
		"gateway": {
			Type:        schema.TypeString,
			Optional:    true,
			Computed:    true,
			Description: "External network gateway IP address",
		},
		"dns": {
			Type:     schema.TypeList,
			Optional: true,
			Computed: true,
			Elem: &schema.Schema{
				Type: schema.TypeString,
			},
			Description: "List of DNS addresses",
		},
		"ntp": {
			Type:     schema.TypeList,
			Optional: true,
			Elem: &schema.Schema{
				Type: schema.TypeString,
			},
			Description: "List of NTP addresses",
		},
		"check_ips": {
			Type:     schema.TypeList,
			Optional: true,
			Computed: true,
			Elem: &schema.Schema{
				Type: schema.TypeString,
			},
			Description: "IPs to check network availability",
		},
		"virtual": {
			Type:        schema.TypeBool,
			Optional:    true,
			Description: "If true - platform DHCP server will not be created",
		},
		"desc": {
			Type:        schema.TypeString,
			Optional:    true,
			Computed:    true,
			Description: "Optional description",
		},
		"start_ip": {
			Type:        schema.TypeString,
			Optional:    true,
			Description: "Start of IP range to be explicitly included",
		},
		"end_ip": {
			Type:        schema.TypeString,
			Optional:    true,
			Description: "End of IP range to be explicitly included",
		},
		"vnfdev_ip": {
			Type:        schema.TypeString,
			Optional:    true,
			Description: "IP to create VNFDev with",
		},
		"pre_reservations_num": {
			Type:        schema.TypeInt,
			Optional:    true,
			Computed:    true,
			Description: "Number of pre created reservations",
		},
		"ovs_bridge": {
			Type:        schema.TypeString,
			Optional:    true,
			Computed:    true,
			Description: "OpenvSwith bridge name for ExtNet connection",
		},
		"enable": {
			Type:        schema.TypeBool,
			Optional:    true,
			Default:     true,
			Description: "Disable/Enable extnet",
		},
		"set_default": {
			Type:        schema.TypeBool,
			Optional:    true,
			Default:     false,
			Description: "Set current extnet as default (can not be undone)",
		},
		"excluded_ips": {
			Type:     schema.TypeSet,
			Optional: true,
			Elem: &schema.Schema{
				Type: schema.TypeString,
			},
			Description: "IPs to exclude in current extnet pool",
		},
		"default_qos": {
			Type:     schema.TypeList,
			MaxItems: 1,
			Optional: true,
			Computed: true,
			Elem: &schema.Resource{
				Schema: map[string]*schema.Schema{
					"e_rate": {
						Type:     schema.TypeInt,
						Optional: true,
						Computed: true,
					},
					"guid": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"in_burst": {
						Type:     schema.TypeInt,
						Optional: true,
						Computed: true,
					},
					"in_rate": {
						Type:     schema.TypeInt,
						Optional: true,
						Computed: true,
					},
				},
			},
		},
		"ckey": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"meta": {
			Type:     schema.TypeList,
			Computed: true,
			Elem: &schema.Schema{
				Type: schema.TypeString,
			},
			Description: "meta",
		},
		"default": {
			Type:     schema.TypeBool,
			Computed: true,
		},
		"free_ips": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"guid": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"extnet_id": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"milestones": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"network_id": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"pri_vnfdev_id": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"shared_with": {
			Type:     schema.TypeList,
			Computed: true,
			Elem: &schema.Schema{
				Type: schema.TypeInt,
			},
		},
		"status": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"vnfs": {
			Type:     schema.TypeList,
			Computed: true,
			Elem: &schema.Resource{
				Schema: map[string]*schema.Schema{
					"dhcp": {
						Type:     schema.TypeInt,
						Computed: true,
					},
				},
			},
		},
		"excluded": {
			Type:     schema.TypeList,
			Computed: true,
			Elem: &schema.Resource{
				Schema: map[string]*schema.Schema{
					"client_type": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"domain_name": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"hostname": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"ip": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"mac": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"type": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"vm_id": {
						Type:     schema.TypeInt,
						Computed: true,
					},
				},
			},
		},
		"network": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"prefix": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"reservations": {
			Type:     schema.TypeList,
			Computed: true,
			Elem: &schema.Resource{
				Schema: map[string]*schema.Schema{
					"client_type": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"domain_name": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"hostname": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"desc": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"ip": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"mac": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"type": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"vm_id": {
						Type:     schema.TypeInt,
						Computed: true,
					},
				},
			},
		},
	}
}

func ResourceExtnetCB() *schema.Resource {
	return &schema.Resource{
		SchemaVersion: 1,

		ReadContext:   resourceExtnetRead,
		CreateContext: resourceExtnetCreate,
		UpdateContext: resourceExtnetUpdate,
		DeleteContext: resourceExtnetDelete,

		Timeouts: &schema.ResourceTimeout{
			Read:    &constants.Timeout300s,
			Create:  &constants.Timeout300s,
			Update:  &constants.Timeout300s,
			Delete:  &constants.Timeout300s,
			Default: &constants.Timeout300s,
		},

		Schema: resourceExtnetSchemaMake(),
	}
}
