/*
Copyright (c) 2019-2022 Digital Energy Cloud Solutions LLC. All Rights Reserved.
Authors:
Petr Krutov, <petr.krutov@digitalenergy.online>
Stanislav Solovev, <spsolovev@digitalenergy.online>
Kasim Baybikov, <kmbaybikov@basistech.ru>
Tim Tkachev, <tvtkachev@basistech.ru>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
Terraform DECORT provider - manage resources provided by DECORT (Digital Energy Cloud
Orchestration Technology) with Terraform by Hashicorp.

Source code: https://repository.basistech.ru/BASIS/terraform-provider-decort

Please see README.md to learn where to place source code so that it
builds seamlessly.

Documentation: https://repository.basistech.ru/BASIS/terraform-provider-decort/wiki
*/

package extnet

import (
	"context"
	"strconv"

	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/schema"
	log "github.com/sirupsen/logrus"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/extnet"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/controller"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/status"
)

func handleExcludedIPsUpdate(ctx context.Context, d *schema.ResourceData, c *controller.ControllerCfg, recNet *extnet.RecordExtNet) error {
	old_set, new_set := d.GetChange("excluded_ips")

	detach_set := old_set.(*schema.Set).Difference(new_set.(*schema.Set))
	if detach_set.Len() > 0 {
		ips := make([]string, 0)
		for _, detach_ip := range detach_set.List() {
			ips = append(ips, detach_ip.(string))
		}

		log.Debugf("cloudbroker: removing %d IP address(es) from excluded list", detach_set.Len())
		req := extnet.IPsIncludeRequest{
			NetID: recNet.ID,
			IPs:   ips,
		}

		_, err := c.CloudBroker().ExtNet().IPsInclude(ctx, req)
		if err != nil {
			return err
		}
	}

	attach_set := new_set.(*schema.Set).Difference(old_set.(*schema.Set))
	if attach_set.Len() > 0 {
		ips := make([]string, 0)
		for _, attach_ip := range attach_set.List() {
			ips = append(ips, attach_ip.(string))
		}

		log.Debugf("cloudbroker: excluding %d IP address(es) from extnet with id %d", attach_set.Len(), recNet.ID)
		req := extnet.IPsExcludeRequest{
			NetID: recNet.ID,
			IPs:   ips,
		}

		_, err := c.CloudBroker().ExtNet().IPsExclude(ctx, req)
		if err != nil {
			return err
		}
	}

	return nil
}

func handleSetDefault(ctx context.Context, d *schema.ResourceData, c *controller.ControllerCfg, recNet *extnet.RecordExtNet) error {
	set_default := d.Get("set_default").(bool)
	if set_default && recNet.Default == false {
		req := extnet.SetDefaultRequest{
			NetID: recNet.ID,
		}

		_, err := c.CloudBroker().ExtNet().SetDefault(ctx, req)
		if err != nil {
			return err
		}
	}

	return nil
}

func handleBasicUpdate(ctx context.Context, d *schema.ResourceData, c *controller.ControllerCfg, recNet *extnet.RecordExtNet) error {
	doBasicUpdate := false
	basiUpdateReq := extnet.UpdateRequest{NetID: recNet.ID}

	if d.HasChange("name") {
		basiUpdateReq.Name = d.Get("name").(string)
		doBasicUpdate = true
	}
	if d.HasChange("desc") {
		basiUpdateReq.Description = d.Get("desc").(string)
		doBasicUpdate = true
	}

	if doBasicUpdate {
		_, err := c.CloudBroker().ExtNet().Update(ctx, basiUpdateReq)
		if err != nil {
			return err
		}
	}

	return nil
}

func handleEnableUpdate(ctx context.Context, d *schema.ResourceData, c *controller.ControllerCfg, recNet *extnet.RecordExtNet) error {
	enable := d.Get("enable").(bool)
	if enable {
		if recNet.Status == status.Disabled {
			req := extnet.EnableRequest{NetID: recNet.ID}
			_, err := c.CloudBroker().ExtNet().Enable(ctx, req)
			if err != nil {
				return err
			}
		}
	} else {
		if recNet.Status == status.Enabled {
			req := extnet.DisableRequest{NetID: recNet.ID}
			_, err := c.CloudBroker().ExtNet().Disable(ctx, req)
			if err != nil {
				return err
			}
		}
	}
	return nil
}

func handleDefaultQOSUpdate(ctx context.Context, d *schema.ResourceData, c *controller.ControllerCfg, recNet *extnet.RecordExtNet) error {
	qos := d.Get("default_qos").([]interface{})[0].(map[string]interface{})

	req := extnet.DefaultQOSUpdateRequest{
		NetID:        recNet.ID,
		IngressRate:  uint64(qos["in_rate"].(int)),
		IngressBurst: uint64(qos["in_burst"].(int)),
		EgressRate:   uint64(qos["e_rate"].(int)),
	}

	_, err := c.CloudBroker().ExtNet().DefaultQOSUpdate(ctx, req)
	if err != nil {
		return err
	}

	return nil
}

func handleNTPUpdate(ctx context.Context, d *schema.ResourceData, c *controller.ControllerCfg, recNet *extnet.RecordExtNet) error {
	changed_list := d.Get("ntp").([]interface{})

	ntp_list := make([]string, 0)
	for _, ntp_address := range changed_list {
		ntp_list = append(ntp_list, ntp_address.(string))
	}

	req := extnet.NTPApplyRequest{
		NetID:   recNet.ID,
		NTPList: ntp_list,
	}

	_, err := c.CloudBroker().ExtNet().NTPApply(ctx, req)
	if err != nil {
		return err
	}

	return nil
}

func handleDNSUpdate(ctx context.Context, d *schema.ResourceData, c *controller.ControllerCfg, recNet *extnet.RecordExtNet) error {
	changed_list := d.Get("dns").([]interface{})

	dns_list := make([]string, 0)
	for _, dns_address := range changed_list {
		dns_list = append(dns_list, dns_address.(string))
	}

	req := extnet.DNSApplyRequest{
		NetID:   recNet.ID,
		DNSList: dns_list,
	}

	_, err := c.CloudBroker().ExtNet().DNSApply(ctx, req)
	if err != nil {
		return err
	}

	return nil
}

func utilityExtnetDefaultCheckPresence(ctx context.Context, m interface{}) (uint64, error) {
	c := m.(*controller.ControllerCfg)

	return c.CloudBroker().ExtNet().GetDefault(ctx)
}

func utilityExtnetListCheckPresence(ctx context.Context, d *schema.ResourceData, m interface{}) (*extnet.ListExtNet, error) {
	c := m.(*controller.ControllerCfg)

	req := extnet.ListRequest{}

	if accountId, ok := d.GetOk("account_id"); ok {
		req.AccountID = uint64(accountId.(int))
	}

	if by_id, ok := d.GetOk("by_id"); ok {
		req.ByID = uint64(by_id.(int))
	}

	if name, ok := d.GetOk("name"); ok {
		req.Name = name.(string)
	}

	if network, ok := d.GetOk("network"); ok {
		req.Network = network.(string)
	}

	if vlan_id, ok := d.GetOk("vlan_id"); ok {
		req.VLANID = uint64(vlan_id.(int))
	}

	if vnfdev_id, ok := d.GetOk("vnfdev_id"); ok {
		req.VNFDevID = uint64(vnfdev_id.(int))
	}

	if status, ok := d.GetOk("status"); ok {
		req.Status = status.(string)
	}

	if page, ok := d.GetOk("page"); ok {
		req.Page = uint64(page.(int))
	}

	if size, ok := d.GetOk("size"); ok {
		req.Size = uint64(size.(int))
	}

	res, err := c.CloudBroker().ExtNet().List(ctx, req)
	if err != nil {
		return nil, err
	}

	return res, nil
}

func utilityExtnetCheckPresence(ctx context.Context, d *schema.ResourceData, m interface{}) (*extnet.RecordExtNet, error) {
	c := m.(*controller.ControllerCfg)
	var netId uint64

	if id, ok := d.GetOk("net_id"); ok {
		netId = uint64(id.(int))
	} else {
		parsed, _ := strconv.ParseUint(d.Id(), 10, 64)
		netId = parsed
	}

	req := extnet.GetRequest{
		NetID: netId,
	}

	res, err := c.CloudBroker().ExtNet().Get(ctx, req)
	if err != nil {
		return nil, err
	}

	return res, nil
}
