/*
Copyright (c) 2019-2022 Digital Energy Cloud Solutions LLC. All Rights Reserved.
Authors:
Petr Krutov, <petr.krutov@digitalenergy.online>
Stanislav Solovev, <spsolovev@digitalenergy.online>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
Terraform DECORT provider - manage resources provided by DECORT (Digital Energy Cloud
Orchestration Technology) with Terraform by Hashicorp.

Source code: https://repository.basistech.ru/BASIS/terraform-provider-decort

Please see README.md to learn where to place source code so that it
builds seamlessly.

Documentation: https://repository.basistech.ru/BASIS/terraform-provider-decort/wiki
*/

package image

import (
	"context"
	"strconv"

	"github.com/hashicorp/terraform-plugin-sdk/v2/diag"
	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/schema"
	log "github.com/sirupsen/logrus"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/image"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/constants"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/controller"
)

func resourceImageCreate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceImageCreate: called for image %s", d.Get("name").(string))

	c := m.(*controller.ControllerCfg)
	req := image.CreateRequest{
		Name:      d.Get("name").(string),
		URL:       d.Get("url").(string),
		GID:       uint64(d.Get("gid").(int)),
		BootType:  d.Get("boot_type").(string),
		ImageType: d.Get("image_type").(string),
	}

	drivers := []string{}
	for _, driver := range d.Get("drivers").([]interface{}) {
		drivers = append(drivers, driver.(string))
	}

	req.Drivers = drivers

	if hotresize, ok := d.GetOk("hot_resize"); ok {
		req.HotResize = hotresize.(bool)
	}
	if username, ok := d.GetOk("username"); ok {
		req.Username = username.(string)
	}
	if password, ok := d.GetOk("password"); ok {
		req.Password = password.(string)
	}
	if accountId, ok := d.GetOk("account_id"); ok {
		req.AccountID = uint64(accountId.(int))
	}
	if usernameDL, ok := d.GetOk("username_dl"); ok {
		req.UsernameDL = usernameDL.(string)
	}
	if passwordDL, ok := d.GetOk("password_dl"); ok {
		req.PasswordDL = passwordDL.(string)
	}
	if sepId, ok := d.GetOk("sep_id"); ok {
		req.SEPID = uint64(sepId.(int))
	}
	if poolName, ok := d.GetOk("pool_name"); ok {
		req.PoolName = poolName.(string)
	}
	if architecture, ok := d.GetOk("architecture"); ok {
		req.Architecture = architecture.(string)
	}

	imageId, err := c.CloudBroker().Image().CreateImage(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	d.SetId(strconv.FormatUint(imageId, 10))
	d.Set("image_id", imageId)

	return resourceImageRead(ctx, d, m)
}

func resourceImageRead(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceImageRead: called for %s id: %s", d.Get("name").(string), d.Id())

	image, err := utilityImageCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	flattenImage(d, image)

	return nil
}

func resourceImageDelete(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceImageDelete: called for %s, id: %s", d.Get("name").(string), d.Id())

	_, err := utilityImageCheckPresence(ctx, d, m)
	if err != nil {
		return diag.FromErr(err)
	}

	c := m.(*controller.ControllerCfg)
	req := image.DeleteRequest{
		ImageID: uint64(d.Get("image_id").(int)),
	}

	if reason, ok := d.GetOk("reason"); ok {
		req.Reason = reason.(string)
	}
	if permanently, ok := d.GetOk("permanently"); ok {
		req.Permanently = permanently.(bool)
	}

	_, err = c.CloudBroker().Image().Delete(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	d.SetId("")

	return nil
}

func resourceImageEditName(ctx context.Context, d *schema.ResourceData, m interface{}) error {
	log.Debugf("resourceImageEditName: called for %s, id: %s", d.Get("name").(string), d.Id())
	c := m.(*controller.ControllerCfg)
	req := image.RenameRequest{
		ImageID: uint64(d.Get("image_id").(int)),
		Name:    d.Get("name").(string),
	}

	_, err := c.CloudBroker().Image().Rename(ctx, req)
	if err != nil {
		return err
	}

	return nil
}

func resourceImageUpdate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceImageEdit: called for %s, id: %s", d.Get("name").(string), d.Id())
	c := m.(*controller.ControllerCfg)

	if d.HasChange("enabled") {
		err := resourceImageChangeEnabled(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("name") {
		err := resourceImageEditName(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("shared_with") {
		err := resourceImageShare(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("computeci_id") {
		err := resourceImageChangeComputeci(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("enabled_stacks") {
		err := resourceImageUpdateNodes(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("link_to") {
		err := resourceImageLink(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChanges("name", "username", "password", "account_id", "bootable", "hot_resize") {
		req := image.EditRequest{}

		req.ImageID = uint64(d.Get("image_id").(int))
		req.Name = d.Get("name").(string)

		req.Username = d.Get("username").(string)
		req.Password = d.Get("password").(string)
		req.AccountID = uint64(d.Get("account_id").(int))
		req.Bootable = d.Get("bootable").(bool)
		req.HotResize = d.Get("hot_resize").(bool)

		_, err := c.CloudBroker().Image().Edit(ctx, req)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	return nil
}

func resourceImageChangeEnabled(ctx context.Context, d *schema.ResourceData, m interface{}) error {
	c := m.(*controller.ControllerCfg)
	imageId := uint64(d.Get("image_id").(int))

	if d.Get("enabled").(bool) {
		req := image.EnableRequest{
			ImageID: imageId,
		}

		_, err := c.CloudBroker().Image().Enable(ctx, req)
		if err != nil {
			return err
		}
	} else {
		req := image.DisableRequest{
			ImageID: imageId,
		}

		_, err := c.CloudBroker().Image().Disable(ctx, req)
		if err != nil {
			return err
		}
	}

	return nil
}

func resourceImageLink(ctx context.Context, d *schema.ResourceData, m interface{}) error {
	log.Debugf("resourceVirtualImageLink: called for %s, id: %s", d.Get("name").(string), d.Id())
	c := m.(*controller.ControllerCfg)
	req := image.LinkRequest{
		ImageID:  uint64(d.Get("image_id").(int)),
		TargetID: uint64(d.Get("link_to").(int)),
	}

	_, err := c.CloudBroker().Image().Link(ctx, req)
	if err != nil {
		return err
	}

	return nil
}

func resourceImageShare(ctx context.Context, d *schema.ResourceData, m interface{}) error {
	log.Debugf("resourceImageShare: called for %s, id: %s", d.Get("name").(string), d.Id())
	c := m.(*controller.ControllerCfg)
	req := image.ShareRequest{
		ImageId: uint64(d.Get("image_id").(int)),
	}
	accIds := []uint64{}
	for _, accId := range d.Get("shared_with").([]interface{}) {
		accIds = append(accIds, uint64(accId.(int)))
	}
	req.AccountIDs = accIds

	_, err := c.CloudBroker().Image().Share(ctx, req)
	if err != nil {
		return err
	}

	return nil
}

func resourceImageChangeComputeci(ctx context.Context, d *schema.ResourceData, m interface{}) error {
	c := m.(*controller.ControllerCfg)

	imageId := uint64(d.Get("image_id").(int))
	computeci := uint64(d.Get("computeci_id").(int))

	if computeci == 0 {
		req := image.ComputeCIUnsetRequest{
			ImageID: imageId,
		}

		_, err := c.CloudBroker().Image().ComputeCIUnset(ctx, req)
		if err != nil {
			return err
		}
	} else {
		req := image.ComputeCISetRequest{
			ImageID:     imageId,
			ComputeCIID: computeci,
		}

		_, err := c.CloudBroker().Image().ComputeCISet(ctx, req)
		if err != nil {
			return err
		}
	}

	return nil
}

func resourceImageUpdateNodes(ctx context.Context, d *schema.ResourceData, m interface{}) error {
	log.Debugf("resourceImageUpdateNodes: called for %s, id: %s", d.Get("name").(string), d.Id())
	c := m.(*controller.ControllerCfg)
	req := image.UpdateNodesRequest{
		ImageID: uint64(d.Get("image_id").(int)),
	}
	enabledStacks := []uint64{}
	for _, stack := range d.Get("enabled_stacks").([]interface{}) {
		enabledStacks = append(enabledStacks, uint64(stack.(int)))
	}

	req.EnabledStacks = enabledStacks

	_, err := c.CloudBroker().Image().UpdateNodes(ctx, req)
	if err != nil {
		return err
	}

	return nil
}

func resourceImageSchemaMake() map[string]*schema.Schema {
	return map[string]*schema.Schema{
		"name": {
			Type:        schema.TypeString,
			Required:    true,
			Description: "Name of the rescue disk",
		},
		"url": {
			Type:        schema.TypeString,
			Required:    true,
			Description: "URL where to download media from",
		},
		"gid": {
			Type:        schema.TypeInt,
			Required:    true,
			Description: "grid (platform) ID where this template should be create in",
		},
		"boot_type": {
			Type:        schema.TypeString,
			Required:    true,
			Description: "Boot type of image bios or uefi",
		},
		"image_type": {
			Type:        schema.TypeString,
			Required:    true,
			Description: "Image type linux, windows or other",
		},
		"drivers": {
			Type:     schema.TypeList,
			Required: true,
			Elem: &schema.Schema{
				Type: schema.TypeString,
			},
			Description: "List of types of compute suitable for image. Example: [ \"KVM_X86\" ]",
		},
		"meta": {
			Type:     schema.TypeList,
			Computed: true,
			Elem: &schema.Schema{
				Type: schema.TypeString,
			},
			Description: "meta",
		},
		"hot_resize": {
			Type:        schema.TypeBool,
			Optional:    true,
			Computed:    true,
			Description: "Does this machine supports hot resize",
		},
		"username": {
			Type:        schema.TypeString,
			Optional:    true,
			Computed:    true,
			Description: "Optional username for the image",
		},
		"password": {
			Type:        schema.TypeString,
			Optional:    true,
			Computed:    true,
			Description: "Optional password for the image",
		},
		"account_id": {
			Type:        schema.TypeInt,
			Optional:    true,
			Computed:    true,
			Description: "AccountId to make the image exclusive",
		},
		"username_dl": {
			Type:        schema.TypeString,
			Optional:    true,
			Computed:    true,
			Description: "username for upload binary media",
		},
		"password_dl": {
			Type:        schema.TypeString,
			Optional:    true,
			Computed:    true,
			Description: "password for upload binary media",
		},
		"sep_id": {
			Type:        schema.TypeInt,
			Optional:    true,
			Computed:    true,
			Description: "storage endpoint provider ID",
		},
		"pool_name": {
			Type:        schema.TypeString,
			Optional:    true,
			Computed:    true,
			Description: "pool for image create",
		},
		"architecture": {
			Type:        schema.TypeString,
			Optional:    true,
			Computed:    true,
			Description: "binary architecture of this image, one of X86_64 of PPC64_LE",
		},
		"image_id": {
			Type:        schema.TypeInt,
			Optional:    true,
			Computed:    true,
			Description: "image id",
		},
		"permanently": {
			Type:        schema.TypeBool,
			Optional:    true,
			Computed:    true,
			Description: "Whether to completely delete the image",
		},
		"bootable": {
			Type:        schema.TypeBool,
			Optional:    true,
			Computed:    true,
			Description: "Does this image boot OS",
		},
		"unc_path": {
			Type:        schema.TypeString,
			Computed:    true,
			Description: "unc path",
		},
		"link_to": {
			Type:        schema.TypeInt,
			Computed:    true,
			Description: "",
		},
		"status": {
			Type:        schema.TypeString,
			Computed:    true,
			Description: "status",
		},
		"tech_status": {
			Type:        schema.TypeString,
			Computed:    true,
			Description: "tech atatus",
		},
		"version": {
			Type:        schema.TypeString,
			Computed:    true,
			Description: "version",
		},
		"size": {
			Type:        schema.TypeInt,
			Computed:    true,
			Description: "image size",
		},
		"enabled": {
			Type:     schema.TypeBool,
			Optional: true,
			Computed: true,
		},
		"computeci_id": {
			Type:     schema.TypeInt,
			Optional: true,
			Computed: true,
		},
		"guid": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"milestones": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"provider_name": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"purge_attempts": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"reference_id": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"res_id": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"res_name": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"rescuecd": {
			Type:     schema.TypeBool,
			Computed: true,
		},
		"reason": {
			Type:     schema.TypeString,
			Optional: true,
		},
		"last_modified": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"desc": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"shared_with": {
			Type:     schema.TypeList,
			Optional: true,
			Computed: true,
			Elem: &schema.Schema{
				Type: schema.TypeInt,
			},
		},
		"sync": {
			Type:        schema.TypeBool,
			Optional:    true,
			Default:     false,
			Description: "Create image from a media identified by URL (in synchronous mode)",
		},
		"enabled_stacks": {
			Type:     schema.TypeList,
			Optional: true,
			Elem: &schema.Schema{
				Type: schema.TypeString,
			},
		},
		"history": {
			Type:     schema.TypeList,
			Computed: true,
			Elem: &schema.Resource{
				Schema: map[string]*schema.Schema{
					"guid": {
						Type:     schema.TypeString,
						Computed: true,
					},
					"id": {
						Type:     schema.TypeInt,
						Computed: true,
					},
					"timestamp": {
						Type:     schema.TypeInt,
						Computed: true,
					},
				},
			},
		},
	}
}

func ResourceImage() *schema.Resource {
	return &schema.Resource{
		SchemaVersion: 1,

		CreateContext: resourceImageCreate,
		ReadContext:   resourceImageRead,
		UpdateContext: resourceImageUpdate,
		DeleteContext: resourceImageDelete,

		Importer: &schema.ResourceImporter{
			StateContext: schema.ImportStatePassthroughContext,
		},

		Timeouts: &schema.ResourceTimeout{
			Create:  &constants.Timeout60s,
			Read:    &constants.Timeout30s,
			Update:  &constants.Timeout60s,
			Delete:  &constants.Timeout60s,
			Default: &constants.Timeout60s,
		},

		Schema: resourceImageSchemaMake(),
	}
}
