/*
Copyright (c) 2019-2022 Digital Energy Cloud Solutions LLC. All Rights Reserved.
Authors:
Petr Krutov, <petr.krutov@digitalenergy.online>
Stanislav Solovev, <spsolovev@digitalenergy.online>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
Terraform DECORT provider - manage resources provided by DECORT (Digital Energy Cloud
Orchestration Technology) with Terraform by Hashicorp.

Source code: https://repository.basistech.ru/BASIS/terraform-provider-decort

Please see README.md to learn where to place source code so that it
builds seamlessly.

Documentation: https://repository.basistech.ru/BASIS/terraform-provider-decort/wiki
*/

package k8s

import (
	"context"

	"github.com/hashicorp/terraform-plugin-sdk/v2/diag"
	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/schema"
	log "github.com/sirupsen/logrus"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/k8s"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/constants"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/controller"
)

func resourceK8sWgCreate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceK8sWgCreate: called with k8s id %d", d.Get("k8s_id").(int))

	c := m.(*controller.ControllerCfg)
	req := k8s.WorkersGroupAddRequest{
		K8SID:      uint64(d.Get("k8s_id").(int)),
		Name:       d.Get("name").(string),
		WorkerNum:  uint64(d.Get("num").(int)),
		WorkerCPU:  uint64(d.Get("cpu").(int)),
		WorkerRAM:  uint64(d.Get("ram").(int)),
		WorkerDisk: uint64(d.Get("disk").(int)),
	}

	resp, err := c.CloudBroker().K8S().WorkersGroupAdd(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	d.SetId(resp)

	return nil
}

func resourceK8sWgRead(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceK8sWgRead: called with k8s id %d", d.Get("k8s_id").(int))

	wg, err := utilityK8sWgCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	d.Set("name", wg.Name)
	d.Set("num", wg.Num)
	d.Set("cpu", wg.CPU)
	d.Set("ram", wg.RAM)
	d.Set("disk", wg.Disk)

	return nil
}

func resourceK8sWgUpdate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceK8sWgUpdate: called with k8s id %d", d.Get("k8s_id").(int))

	c := m.(*controller.ControllerCfg)

	wg, err := utilityK8sWgCheckPresence(ctx, d, m)
	if err != nil {
		return diag.FromErr(err)
	}

	addReq := k8s.WorkerAddRequest{
		K8SID:          uint64(d.Get("k8s_id").(int)),
		WorkersGroupID: wg.ID,
	}

	delReq := k8s.DeleteWorkerFromGroupRequest{
		K8SID:          uint64(d.Get("k8s_id").(int)),
		WorkersGroupID: wg.ID,
	}

	if newNum := uint64(d.Get("num").(int)); newNum > wg.Num {
		addReq.Num = newNum - wg.Num
		_, err := c.CloudBroker().K8S().WorkerAdd(ctx, addReq)
		if err != nil {
			return diag.FromErr(err)
		}
	} else {
		for i := wg.Num - 1; i >= newNum; i-- {
			delReq.WorkerID = wg.DetailedInfo[i].ID
			_, err := c.CloudBroker().K8S().DeleteWorkerFromGroup(ctx, delReq)
			if err != nil {
				return diag.FromErr(err)
			}
		}
	}

	return nil
}

func resourceK8sWgDelete(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceK8sWgDelete: called with k8s id %d", d.Get("k8s_id").(int))

	wg, err := utilityK8sWgCheckPresence(ctx, d, m)
	if err != nil {
		return diag.FromErr(err)
	}

	c := m.(*controller.ControllerCfg)
	req := k8s.WorkersGroupDeleteRequest{
		K8SID:          uint64(d.Get("k8s_id").(int)),
		WorkersGroupID: wg.ID,
	}

	_, err = c.CloudBroker().K8S().WorkersGroupDelete(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	d.SetId("")

	return nil
}

func resourceK8sWgSchemaMake() map[string]*schema.Schema {
	return map[string]*schema.Schema{
		"k8s_id": {
			Type:        schema.TypeInt,
			Required:    true,
			ForceNew:    true,
			Description: "ID of k8s instance.",
		},

		"name": {
			Type:        schema.TypeString,
			Required:    true,
			ForceNew:    true,
			Description: "Name of the worker group.",
		},

		"num": {
			Type:        schema.TypeInt,
			Optional:    true,
			Default:     1,
			Description: "Number of worker nodes to create.",
		},

		"cpu": {
			Type:        schema.TypeInt,
			Optional:    true,
			ForceNew:    true,
			Default:     1,
			Description: "Worker node CPU count.",
		},

		"ram": {
			Type:        schema.TypeInt,
			Optional:    true,
			ForceNew:    true,
			Default:     1024,
			Description: "Worker node RAM in MB.",
		},

		"disk": {
			Type:        schema.TypeInt,
			Optional:    true,
			ForceNew:    true,
			Default:     0,
			Description: "Worker node boot disk size. If unspecified or 0, size is defined by OS image size.",
		},
	}
}

func ResourceK8sWg() *schema.Resource {
	return &schema.Resource{
		SchemaVersion: 1,

		CreateContext: resourceK8sWgCreate,
		ReadContext:   resourceK8sWgRead,
		UpdateContext: resourceK8sWgUpdate,
		DeleteContext: resourceK8sWgDelete,

		Importer: &schema.ResourceImporter{
			StateContext: schema.ImportStatePassthroughContext,
		},

		Timeouts: &schema.ResourceTimeout{
			Create:  &constants.Timeout20m,
			Read:    &constants.Timeout30s,
			Update:  &constants.Timeout20m,
			Delete:  &constants.Timeout60s,
			Default: &constants.Timeout60s,
		},

		Schema: resourceK8sWgSchemaMake(),
	}
}
