/*
Copyright (c) 2019-2022 Digital Energy Cloud Solutions LLC. All Rights Reserved.
Authors:
Petr Krutov, <petr.krutov@digitalenergy.online>
Stanislav Solovev, <spsolovev@digitalenergy.online>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
Terraform DECORT provider - manage resources provided by DECORT (Digital Energy Cloud
Orchestration Technology) with Terraform by Hashicorp.

Source code: https://repository.basistech.ru/BASIS/terraform-provider-decort

Please see README.md to learn where to place source code so that it
builds seamlessly.

Documentation: https://repository.basistech.ru/BASIS/terraform-provider-decort/wiki
*/

package kvmvm

import (
	"context"
	"fmt"
	"strconv"

	log "github.com/sirupsen/logrus"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/compute"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/rg"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/controller"

	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/schema"
)

func utilityComputeExtraDisksConfigure(ctx context.Context, d *schema.ResourceData, m interface{}, do_delta bool) error {
	c := m.(*controller.ControllerCfg)
	computeID, _ := strconv.ParseUint(d.Id(), 10, 64)

	log.Debugf("utilityComputeExtraDisksConfigure: called for Compute ID %s with do_delta = %t", d.Id(), do_delta)

	old_set, new_set := d.GetChange("extra_disks")

	apiErrCount := 0
	var lastSavedError error

	if !do_delta {
		if new_set.(*schema.Set).Len() < 1 {
			return nil
		}

		for _, disk := range new_set.(*schema.Set).List() {
			req := compute.DiskAttachRequest{
				ComputeID: computeID,
				DiskID:    uint64(disk.(int)),
			}

			_, err := c.CloudBroker().Compute().DiskAttach(ctx, req)
			if err != nil {
				apiErrCount++
				lastSavedError = err
			}
		}

		if apiErrCount > 0 {
			log.Errorf("utilityComputeExtraDisksConfigure: there were %d error(s) when attaching disks to Compute ID %s. Last error was: %s",
				apiErrCount, d.Id(), lastSavedError)
			return lastSavedError
		}

		return nil
	}

	detach_set := old_set.(*schema.Set).Difference(new_set.(*schema.Set))
	log.Debugf("utilityComputeExtraDisksConfigure: detach set has %d items for Compute ID %s", detach_set.Len(), d.Id())
	for _, diskId := range detach_set.List() {
		req := compute.DiskDetachRequest{
			ComputeID: computeID,
			DiskID:    uint64(diskId.(int)),
		}

		_, err := c.CloudBroker().Compute().DiskDetach(ctx, req)
		if err != nil {
			// failed to detach disk - there will be partial resource update
			log.Errorf("utilityComputeExtraDisksConfigure: failed to detach disk ID %d from Compute ID %s: %s", diskId.(int), d.Id(), err)
			apiErrCount++
			lastSavedError = err
		}
	}

	attach_set := new_set.(*schema.Set).Difference(old_set.(*schema.Set))
	log.Debugf("utilityComputeExtraDisksConfigure: attach set has %d items for Compute ID %s", attach_set.Len(), d.Id())
	for _, diskId := range attach_set.List() {
		req := compute.DiskAttachRequest{
			ComputeID: computeID,
			DiskID:    uint64(diskId.(int)),
		}

		_, err := c.CloudBroker().Compute().DiskAttach(ctx, req)
		if err != nil {
			// failed to attach disk - there will be partial resource update
			log.Errorf("utilityComputeExtraDisksConfigure: failed to attach disk ID %d to Compute ID %s: %s", diskId.(int), d.Id(), err)
			apiErrCount++
			lastSavedError = err
		}
	}

	if apiErrCount > 0 {
		log.Errorf("utilityComputeExtraDisksConfigure: there were %d error(s) when managing disks of Compute ID %s. Last error was: %s",
			apiErrCount, d.Id(), lastSavedError)
		return lastSavedError
	}

	return nil
}

func utilityComputeNetworksConfigure(ctx context.Context, d *schema.ResourceData, m interface{}, do_delta bool) error {
	c := m.(*controller.ControllerCfg)
	computeID, _ := strconv.ParseUint(d.Id(), 10, 64)

	old_set, new_set := d.GetChange("network")

	apiErrCount := 0
	var lastSavedError error

	if !do_delta {
		if new_set.(*schema.Set).Len() < 1 {
			return nil
		}

		for _, runner := range new_set.(*schema.Set).List() {
			net_data := runner.(map[string]interface{})
			req := compute.NetAttachRequest{
				ComputeID: computeID,
				NetType:   net_data["net_type"].(string),
				NetID:     uint64(net_data["net_id"].(int)),
			}

			ipaddr, ipSet := net_data["ip_address"]
			if ipSet {
				req.IPAddr = ipaddr.(string)
			}
			_, err := c.CloudBroker().Compute().NetAttach(ctx, req)
			if err != nil {
				apiErrCount++
				lastSavedError = err
			}
		}

		if apiErrCount > 0 {
			log.Errorf("utilityComputeNetworksConfigure: there were %d error(s) when managing networks of Compute ID %s. Last error was: %s",
				apiErrCount, d.Id(), lastSavedError)
			return lastSavedError
		}
		return nil
	}

	detach_set := old_set.(*schema.Set).Difference(new_set.(*schema.Set))
	log.Debugf("utilityComputeNetworksConfigure: detach set has %d items for Compute ID %s", detach_set.Len(), d.Id())
	for _, runner := range detach_set.List() {
		net_data := runner.(map[string]interface{})
		req := compute.NetDetachRequest{
			ComputeID: computeID,
			IPAddr:    net_data["ip_address"].(string),
			MAC:       net_data["mac"].(string),
		}

		_, err := c.CloudBroker().Compute().NetDetach(ctx, req)
		if err != nil {
			// failed to detach this network - there will be partial resource update
			log.Errorf("utilityComputeNetworksConfigure: failed to detach net ID %d of type %s from Compute ID %s: %s",
				net_data["net_id"].(int), net_data["net_type"].(string), d.Id(), err)
			apiErrCount++
			lastSavedError = err
		}
	}

	attach_set := new_set.(*schema.Set).Difference(old_set.(*schema.Set))
	log.Debugf("utilityComputeNetworksConfigure: attach set has %d items for Compute ID %s", attach_set.Len(), d.Id())
	for _, runner := range attach_set.List() {
		net_data := runner.(map[string]interface{})
		req := compute.NetAttachRequest{
			ComputeID: computeID,
			NetID:     uint64(net_data["net_id"].(int)),
			NetType:   net_data["net_type"].(string),
		}

		if net_data["ip_address"].(string) != "" {
			req.IPAddr = net_data["ip_address"].(string)
		}

		_, err := c.CloudBroker().Compute().NetAttach(ctx, req)
		if err != nil {
			// failed to attach this network - there will be partial resource update
			log.Errorf("utilityComputeNetworksConfigure: failed to attach net ID %d of type %s to Compute ID %s: %s",
				net_data["net_id"].(int), net_data["net_type"].(string), d.Id(), err)
			apiErrCount++
			lastSavedError = err
		}
	}

	if apiErrCount > 0 {
		log.Errorf("utilityComputeNetworksConfigure: there were %d error(s) when managing networks of Compute ID %s. Last error was: %s",
			apiErrCount, d.Id(), lastSavedError)
		return lastSavedError
	}

	return nil
}

func utilityComputeCheckPresence(ctx context.Context, d *schema.ResourceData, m interface{}) (*compute.RecordCompute, error) {
	c := m.(*controller.ControllerCfg)

	idSet := false
	computeID, err := strconv.ParseUint(d.Id(), 10, 64)
	if err != nil || computeID <= 0 {
		computeId, argSet := d.GetOk("compute_id") // NB: compute_id is NOT present in computeResource schema!
		if argSet {
			computeID = uint64(computeId.(int))
			idSet = true
		}
	} else {
		idSet = true
	}

	if idSet {
		// compute ID is specified, try to get compute instance straight by this ID
		log.Debugf("utilityComputeCheckPresence: locating compute by its ID %d", computeID)
		req := compute.GetRequest{
			ComputeID: computeID,
		}

		computeFacts, err := c.CloudBroker().Compute().Get(ctx, req)
		if err != nil {
			return nil, err
		}
		return computeFacts, nil
	}

	// ID was not set in the schema upon entering this function - work through Compute name
	// and RG ID
	computeName, argSet := d.GetOk("name")
	if !argSet {
		return nil, fmt.Errorf("Cannot locate compute instance if name is empty and no compute ID specified")
	}

	rgId, argSet := d.GetOk("rg_id")
	if !argSet {
		return nil, fmt.Errorf("Cannot locate compute by name %s if no resource group ID is set", computeName.(string))
	}

	rgListComputesReq := rg.ListComputesRequest{
		RGID: uint64(rgId.(int)),
	}

	computeList, err := c.CloudBroker().RG().ListComputes(ctx, rgListComputesReq)
	if err != nil {
		return nil, err
	}

	log.Debugf("utilityComputeCheckPresence: traversing decoded JSON of length %d", len(computeList.Data))
	for index, item := range computeList.Data {
		// need to match Compute by name, skip Computes with the same name in DESTROYED satus
		if item.Name == computeName.(string) && item.Status != "DESTROYED" {
			log.Debugf("utilityComputeCheckPresence: index %d, matched name %s", index, item.Name)
			// we found the Compute we need - now get detailed information via compute/get API
			req := compute.GetRequest{
				ComputeID: item.ID,
			}

			apiResp, err := c.CloudBroker().Compute().Get(ctx, req)
			if err != nil {
				return nil, err
			}
			return apiResp, nil
		}
	}

	return nil, nil
}
